"""
Business logic for Vulture's mock logging, jobs, and dags
"""

import logging
import re
import subprocess
import sys
from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List, NamedTuple, Tuple, Union

import pendulum
from dag import DAG

from workspaces.schema import WorkflowEventType


class MockLogger:
    """
    Class representing a logger for Vulture that will simulate HTCondor logs
    """

    def __init__(self, log_path: Path, out: str, err: str):
        self.out = out
        self.err = err
        self.logger = logging.getLogger("vulture")

        self.logger.setLevel(logging.INFO)
        self.def_handler = logging.FileHandler(f"{log_path}", mode="a+")
        self.logger.addHandler(self.def_handler)

    @contextmanager
    def switch_logger_file(self, log_file: str):
        """
        Switches logger FileHandler to write to a new file; operates as a context manager

        :param log_file: File name of the log you want to switch to
        """
        new_handler = logging.FileHandler(log_file, mode="w+")
        self.logger.removeHandler(self.def_handler)
        self.logger.addHandler(new_handler)
        yield
        self.logger.removeHandler(new_handler)
        self.logger.addHandler(self.def_handler)

    def write_out(self, which_out: str, output: str):
        """
        Writes output to an auxiliary log file (any file that is not the main condor.log)

        :param which_out: Which log is to be written to (out/error)
        :param output: Contents of the file
        :raises ValueError: If which_out is not a supported log type
        """
        # hello, world
        if which_out == "out":
            with self.switch_logger_file(self.out):
                self.logger.info(output)
        elif which_out == "err":
            with self.switch_logger_file(self.err):
                self.logger.info(output)
        else:
            # If adding new logging option, add to supported list
            supported = ["out", "err"]
            raise ValueError(
                f"Output '{which_out}' not supported. Use [{'|'.join(supported)}]."
            )

    def log_event(self, event_type: WorkflowEventType, event_body: str = None):
        """
        Log a mock HTCondor event to the log file

        :param event_type: Type of the event
        :param event_body: Body of the event log (can be empty)
        """
        header = self.create_header(event_type)
        # Indent each line of the body
        if event_body:
            event_body_list = event_body.split("\n")
            if len(event_body_list) > 1:
                event_body = "\t".join(event_body_list)
            else:
                event_body = "\t" + event_body
        footer = "..."
        self.logger.info(
            msg="\n".join([header, event_body, footer])
        ) if event_body else self.logger.info("\n".join([header, footer]))

    @staticmethod
    def create_header(event_type: WorkflowEventType) -> str:
        """
        Create mock HTCondor event log header

        EXAMPLE:
        000 (3983.000.000) 08/26 11:06:06 Job submitted from host: <10.64.1.178:9618?addrs=10.64.1.178-9618&noUDP&sock=4050180_7e37_3>

        :param event_type: Type of event being logged
        :raises ValueError: If event_type is not yet supported
        :return: The header as a string
        """
        mock_host = "<0.0.0.0:7777>"
        event_descs = {
            WorkflowEventType.SUBMITTED: f"Job submitted from host: {mock_host}",
            WorkflowEventType.EXECUTING: f"Job executing on host: {mock_host}",
            WorkflowEventType.TERMINATED: "Job terminated.",
            WorkflowEventType.OTHER: "Unknown description.",
        }
        event_num = f"{event_type.value:03d}"
        # TODO: Add calculable job number
        job_num = f"({4242:04d}.000.000)"
        timestamp = f"{pendulum.now().format('MM/DD HH:mm:ss')}"
        if event_type in event_descs.keys():
            event_desc = f"{event_descs[event_type]}"
        else:
            # If this gets triggered, add event type to event_descs dict above
            raise ValueError(f"ERROR: Event type {event_type} not supported.")
        return " ".join([event_num, job_num, timestamp, event_desc])


class MalformedJobFile(ValueError):
    pass


class Job:
    """
    Class representing a faked HTCondor job
    """

    def __init__(self, file: Path, write_log: bool = True):
        """
        :param file: Path to job file
        :param write_log: Boolean that determines whether a log file will be written
        """
        self.fields, self.command = self.parse(file)
        self.write_log = write_log
        if self.write_log:
            if self.fields.get("executable"):
                # Collect logging file names from job file if they exist; else give them sane defaults
                exec_name = self.fields["executable"]
                self.logger = MockLogger(
                    log_path=self.fields.get("log", "./condor.log"),
                    out=self.fields.get("output", f"{exec_name}.out"),
                    err=self.fields.get("error", f"{exec_name}.error"),
                )
                # Job is submitted, write an event to signify that
                self.logger.log_event(
                    WorkflowEventType.SUBMITTED,
                    f"Job node: {self.fields['executable']}",
                )
            else:
                raise MalformedJobFile(
                    f"ERROR: Condor job file {file} malformed. No executable found in fields {self.fields}."
                )

    def parse(self, file: Path) -> NamedTuple:
        """
        Parse condor job file for relevant fields
        TODO: Make parsing more robust

        :param file: Path to job file
        :return: NamedTuple of job data
        """
        # Parse file looking for executable and arguments
        JobData = namedtuple("JobData", ["field_data", "command"])
        fields = []

        try:
            with open(file, "r") as f:
                # Split file into fields and command
                for line in f.readlines():
                    if re.search(r" = ", line):
                        fields.append(line.rstrip())
                    else:
                        command = line.rstrip()
        except IOError as e:
            print(f"ERROR: Could not open/read '{file}'.")
            sys.exit(1)

        if len(fields) > 0:
            field_dict = self.parse_fields(fields)
        else:
            raise MalformedJobFile(
                f"ERROR: Condor job file '{file}' is malformed. No fields found."
            )
        return JobData(field_dict, command)

    @staticmethod
    def parse_fields(fields: List[str]) -> Dict[str, str]:
        """
        Take in list of fields parsed from input file and return them as a dictionary

        :param fields: List of fields
        :return: Dictionary of fields
        """
        field_dict = {}
        for field in fields:
            key, value = field.split(" = ")
            field_dict[key] = value
        return field_dict

    def execute(self):
        """
        Execute given process with given arguments
        TODO: Buff return code processing
        """
        result_strings = {
            0: "Normal termination (return value 0)",
            # This is 255 instead of -1 because that's what subprocess returns instead of -1
            255: "Error in execution (return value -1)",
        }
        capture_output = False
        if self.write_log:
            self.logger.log_event(WorkflowEventType.EXECUTING)
            capture_output = True

        # Execute
        if self.fields.get("executable"):
            process = subprocess.run(
                [self.fields["executable"], self.fields["arguments"]],
                capture_output=capture_output,
                text=True,
            )
        else:
            raise MalformedJobFile(
                f"ERROR: Condor job file malformed. No executable found in fields {self.fields}."
            )

        exec_results = f"{result_strings.get(process.returncode, f'Something weird happened (return value {process.returncode})')}"
        if self.write_log:
            self.logger.log_event(WorkflowEventType.TERMINATED, exec_results)
            self.logger.write_out("out", process.stdout)
            self.logger.write_out("err", process.stderr)


class MalformedDagFile(ValueError):
    """
    Exception indicating the given DAG file contains an error
    """

    pass


# Type aliases
JobFileMapping = Dict[str, str]
Parents = List[str]
Children = List[str]


class DAGMan:
    """
    Class representing a faked HTCondor directed acyclic graph; uses py-dag.DAG library (https://github.com/thieman/py-dag)
    """

    def __init__(self, file: Path):
        """
        :param file: Path to HTCondor-DAG-formatted input file (see https://htcondor.readthedocs.io/en/latest/users-manual/dagman-workflows.html
        """
        self.dag, self.files = self.parse(file)

    @staticmethod
    def parse_node(line: str) -> Tuple[str, str]:
        """
        Parse line containing DAG node definition and return node definition

        :param line: Line containing node definition
        :return: DAG node definition (job name, job file)
        """
        # Parse node from line
        node = re.split(r"\s+", line)
        job_name, job_file_name = node[1], node[2]

        return job_name, job_file_name

    @staticmethod
    def parse_edge(line: str) -> Tuple[Parents, Children]:
        """
        Parse line containing DAG edge definition and return edge definition

        :param line: Line containing edge definition
        :return: DAG edge definition (parents, children)
        """
        # Regexes
        r_parents = r"^PARENT\s+(?P<parents>.*?)\s+CHILD"
        r_children = r"CHILD\s+(?P<children>.*?)$"
        # Parse edge from line
        parents = re.split(r"\s+", re.search(r_parents, line).group("parents"))
        children = re.split(r"\s+", re.search(r_children, line).group("children"))
        return parents, children

    def parse_lines(self, lines: List[str]) -> Tuple[DAG, JobFileMapping]:
        """
        Parse a list of lines from the input file into a DAG object

        :param lines: List of lines
        :return: Directed acyclic graph object and dict with job/jobfile associations
        """
        # Initialize empty directed acyclic graph
        dag = DAG()
        # Initialize empty job-file mapping
        files = {}
        # Regexes
        r_job_name = r"\b(?!PARENT)(?!CHILD)[A-z0-9-_]+\b"
        for line in lines:
            if re.search(r"JOB", line):
                # Line is a node definition
                job_name, job_file = self.parse_node(line)
                # Add DAG node
                dag.add_node(job_name)
                # Add file association
                files[job_name] = job_file
            elif re.search(
                rf"PARENT\s+(?:{r_job_name}\s+)+CHILD\s+(?:{r_job_name}\s+)+", line
            ):
                # Line is an edge definition
                # Add edges to DAG
                parents, children = self.parse_edge(line)
                for parent in parents:
                    for child in children:
                        dag.add_edge(parent, child)
            else:
                # Line is not a node or an edge
                pass

        # Error checking
        if len(files) == 0 or dag.size() == 0:
            raise MalformedDagFile(
                "ERROR: DAG was unable to be created. DAG file not formatted correctly."
            )
        if not dag.validate()[0]:
            raise MalformedDagFile(
                "ERROR: DAG created by file is not valid. Is there a cycle?"
            )

        return dag, files

    def parse(self, file: Path) -> DAG:
        """
        Given an HTCondor-formatted DAG file, parse it and convert it to a DAG object

        :param file: DAG file
        :return: Directed acyclic graph object
        """
        try:
            with open(file, "r") as f:
                lines = f.readlines()
        except IOError as e:
            print(f"ERROR: Could not open/read '{file}'.")
            sys.exit(1)

        dag = self.parse_lines(lines)

        return dag

    def execute(self, write_log: bool = True):
        """
        Create jobs for each DAG node and execute them in order

        NOTE: Node execution is not asynchronous for simplicity

        :param write_log: Boolean that determines whether a log file will be written
        """
        dag_nodes_sorted = self.dag.topological_sort()
        for node in dag_nodes_sorted:
            # Iterate through DAG nodes ordered by topological sort
            # Get corresponding job file
            job_file = self.files[node]
            # Create job
            job = Job(Path(job_file), write_log)
            # Execute job
            job.execute()
