"""
Testing suite for wf_monitor. Currently not covered by tests:
- log_decorator_factory
- make_arg_parser
"""

import logging
import os
import threading
import time

import pytest
from pytest_mock import MockerFixture
from wf_monitor.monitor import WorkflowMonitor

from workspaces.schema import WorkflowEventType

os.environ["CAPO_PROFILE"] = "docker"
log_path = "logs/condor.log"
test_monitor = WorkflowMonitor(log_path, -1)
WORKFLOW_STATUS_EXCH = "workspaces.workflow-service.workflow-status"


def test_read_log():
    """
    Tests whether or not the example log is being correctly read by checking for strings known to
    exist in the example log file
    """
    test_strs = [
        "Image size of job updated: 72",
        "72  -  ResidentSetSize of job (KB)",
        "(1) Normal termination (return value 0)",
        "000 (3983.000.000) 08/26 11:06:06 Job submitted from host: "
        "<10.64.1.178:9618?addrs=10.64.1.178-9618&noUDP&sock=4050180_7e37_3>",
    ]
    for s in test_strs:
        assert s in test_monitor.log


def slow_log_helper():
    WorkflowMonitor("logs/slow.log", -1)


@pytest.mark.skip(
    reason="This test sometimes fails. Skipping until it is investigated further."
)
def test_read_log_slow(caplog):
    """
    Tests that wf_monitor can keep up with reading when writing is slow.
    """
    caplog.set_level(logging.INFO)
    thread = threading.Thread(target=slow_log_helper)
    thread.start()
    # Truncate slow.log
    open("logs/slow.log", "w").close()
    event = ""
    with open("logs/condor.log", "r") as log:
        for line in log:
            event += line
            if line == "...\n":
                # End of event reached
                with open("logs/slow.log", "a") as slow:
                    slow.write(event)
                    event = ""
                    time.sleep(1)

    thread.join()
    assert "Finished monitoring logs/slow.log." in caplog.text


def test_read_log_timeout():
    """
    Tests the timeout functionality of WorkflowMonitor.read_htcondor_log()
    """
    with pytest.raises(SystemExit) as sys_ex:
        WorkflowMonitor("logs/file-that-does-not-exist.txt", -1, 1)
    assert sys_ex.type == SystemExit
    assert sys_ex.value.code == -1


def test_parse_log():
    """
    Tests that the example log file is correctly parsed by checking for the known order of event
    types
    """
    test_event_types = [
        WorkflowEventType.SUBMITTED,
        WorkflowEventType.EXECUTING,
        WorkflowEventType.OTHER,
        WorkflowEventType.TERMINATED,
    ]
    for e, e_type in zip(test_monitor.events, test_event_types):
        assert e.type == e_type


def test_parse_log_error(caplog):
    """
    Tests that WorkflowMonitor.parse_log() correctly raises a ValueError when a badly formatted
    log is parsed
    """
    caplog.set_level(logging.INFO)
    with pytest.raises(TimeoutError):
        WorkflowMonitor("logs/test.log", -1, 5)
    msg = (
        "HTCondor event not well-formatted. Event contents:\n"
        "this will trigger an exception!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
        "...\n"
        "\n"
    )
    assert "HTCondor event not well-formatted." in caplog.text


def test_send_events(mocker: MockerFixture):
    """
    Test that mocks the core wf_monitor functionality of creating a connection to the RabbitMQ
    server, configuring a channel with that connection, and emitting events through that channel

    :param mocker: Mocker fixture from pytest_mock
    """
    mock_workflow_events = mocker.patch("channels.amqp_helpers.Channel")

    for event in WorkflowMonitor(log_path, 64).events:
        mock_workflow_events.send(event, WORKFLOW_STATUS_EXCH)

    assert mock_workflow_events.send.call_count == 4
