""" Test data: ALMA download products """

import warnings

from schema import Project, Author, ExecutionBlock, ScienceProduct
from sqlalchemy import exc as sa_exc

from .expected_values_evla import (
    EXECUTION_BLOCK,
    METADATA_INGESTION_DATE,
    CASA_LOG,
    PPR_FILENAME,
)
from shared.workspaces.test.test_data.utilities import get_file_info_from_json_file

EXEC_BLOCK = "execblock"
CALIBRATED = "CALIBRATED"
WEBLOG = "weblog.tgz"
PIPELINE_MANIFEST = "unknown.download_alma_ms.pipeline_manifest.xml"
PIPELINE_AQUAREPORT = "pipeline_aquareport.xml"
CASA_PIPESCRIPT = "casa_pipescript.py"
AUX_PRODUCTS = "unknown.download_alma_ms.auxproducts.tgz"
MS_PP_REQUEST = "unknown.download_alma_ms.pprequest.xml"

# pylint: disable=R0201, R0801, R0902, R0903


class AlmaPublicProductSet:
    """ Test data for ALMA project 2017.1.00297.S """

    def __init__(self):
        self.project = self.build_project()
        self.exec_blocks = self.build_exec_blocks()
        self.science_products = self.build_science_products()
        self.file_info = get_file_info_from_json_file("A001_X1296_Xa93_RAW.json")
        self.rawdata_total_size = self._compute_rawdata_size()
        self.measurement_set, self.basic_ms_size = self.build_basic_ms()
        self.restored_ms, self.rest_ms_size = self.build_restored_ms()

    def build_project(self) -> Project:
        """Returns project metadata for 2017.1.00297.S,
        including authors' details
        """

        with warnings.catch_warnings():
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            project_code = "2017.1.00297.S"
            project = Project(
                project_code=project_code,
                starttime=56799.3877155556,
                endtime=56799.4128683333,
                proprietary_duration=365,
                total_observation_time=0.0583143749972805,
                title="Copy of 2012.1.00060.S for testing",
                abstract="Understanding the physical factors that control the conversion of interstellar gas into stars is of fundamental importance for both developing a predictive physical theory of star formation and understanding the evolution of galaxies from the earliest epochs of cosmic history to the present time. An important aspect of this question is the study of empirical relations that connect the star formation rate in a given region to local properties of the interstellar medium. An important example is the Schmidt-Kennicutt (KS) law for galaxies that relates the surface densities of the star formation rate and the surface densities of interstellar gas in a non-linear fashion. However, it is also known that there is a linear correlation between the total SFR in galaxies and the mass of dense molecular gas as traced by the high excitation HCN molecule. Contrary to the KS relation, this scaling relation suggests that the total SFR depends simply on the total amount of dense molecular gas in a star forming system. Recently, we have begun to test these scaling relations in the Galactic neighborhood where star formation rates can be much better constrained. We found that for local clouds the total SFR scales most directly, and linearly, with the total mass of high extinction (and dense) molecular gas. Furthermore, we found this linear scaling law between SFR and dense gas to extend and extrapolate directly and smoothly to external galaxies. Moreover, our observations also demonstrate that a KS type relation does not exist for molecular clouds in the Galactic neighborhood. This is a direct consequence of a well known scaling law between the mass and size of molecular clouds, Larson's third law. Overall, our results indicate that a linear scaling law, in which the total amount of dense gas controls the SFR, is the fundamental physical relation that connects star formation across the vast scales from individual GMCs to entire galaxies. Critical testing of these ideas require resolved observations of GMCs in external galaxies. Here we propose to use ALMA to evaluate star formation scaling laws in a nearby galaxy where we can obtain resolved observations of individual GMCs. This allows us to obtain observations of a larger sample of GMCs than is accessible in the Galactic neighborhood. An extensive APEX survey of HII regions in the nearby galaxy NGC 300 has provided us with a sample of 36 star-forming regions with CO(2-1) detections and 42 upper limits. We are currently working on obtaining star formation rates for these regions from multi-wavelength ancillary data including our Herschel observations. We propose to use ALMA's unequalled capabilities to obtain snapshot observations of 40 selected regions in CO(2-1) in order to make resolved measurements of cloud structure to obtain sizes and virial masses. As a pilot project, we also propose to observe the brightest subsample in HCN(1-0) as a dense-gas tracer. Our proposed ALMA CO observations will enable us to to test Larson's scaling laws in an external galaxy and to evaluate which formulation of the Schmidt law is the most meaningful and appropriate to apply to spiral galaxies, and in doing so refine Schmidt's original conjecture of a scaling relation between the rate of star formation and gas density.",
            )
            project.authors = [
                Author(
                    project_code=project_code,
                    author_id=32550,
                    username="franzbauer",
                    firstname="Franz",
                    lastname="Bauer",
                    pst_person_id=None,
                    is_pi=True,
                ),
                Author(
                    project_code=project_code,
                    author_id=32551,
                    username="ezetre",
                    firstname="Ezequiel",
                    lastname="Treister",
                    pst_person_id=None,
                    is_pi=False,
                ),
                Author(
                    project_code=project_code,
                    author_id=32552,
                    username="lho",
                    firstname="Luis",
                    lastname="Ho",
                    pst_person_id=None,
                    is_pi=False,
                ),
                Author(
                    project_code=project_code,
                    author_id=32554,
                    username="jyshangguan",
                    firstname="Jinyi",
                    lastname="Shangguan",
                    pst_person_id=None,
                    is_pi=False,
                ),
                Author(
                    project_code=project_code,
                    author_id=32553,
                    username="rwang",
                    firstname="Ran",
                    lastname="Wang",
                    pst_person_id=None,
                    is_pi=False,
                ),
            ]

            return project

    def build_exec_blocks(self) -> list:
        """ Returns execution blocks for OUS uid://A001/X1296/Xa93. """
        alma_ous_id = "uid://A001/X1296/Xa93"
        band_code = "05"

        return [
            ExecutionBlock(
                project_code=self.project.project_code,
                execution_block_id=102209,
                alma_ous_id=alma_ous_id,
                band_code=band_code,
                filegroup_id=212043,
                ngas_fileset_id="uid___A002_Xcd8029_Xa892",
                starttime=58256.0535855556,
                endtime=58256.1137983333,
                calibration_status=CALIBRATED,
            ),
            ExecutionBlock(
                project_code=self.project.project_code,
                execution_block_id=102169,
                alma_ous_id=alma_ous_id,
                band_code=band_code,
                filegroup_id=211958,
                ngas_fileset_id="uid___A002_Xcd8029_Xfdd",
                starttime=58255.0338461111,
                endtime=58255.0919594444,
                calibration_status=CALIBRATED,
            ),
            ExecutionBlock(
                project_code=self.project.project_code,
                execution_block_id=102211,
                alma_ous_id=alma_ous_id,
                band_code=band_code,
                filegroup_id=212047,
                ngas_fileset_id="uid___A002_Xcd8029_Xb0a4",
                starttime=58256.1152683333,
                endtime=58256.1753838889,
                calibration_status=CALIBRATED,
            ),
        ]

    def build_science_products(self) -> list:
        """ Returns science products assocated with project """

        external_system = "ALMA Operations"
        return [
            ScienceProduct(
                external_name="uid___A002_Xcd8029_Xfdd",
                filegroup_id=211958,
                external_system=external_system,
                metadata_ingestion_date=METADATA_INGESTION_DATE,
                metadata_ingestion_version="1",
                science_product_type=EXECUTION_BLOCK,
            ),
            ScienceProduct(
                external_name="uid___A002_Xcd8029_Xb0a4",
                filegroup_id=212047,
                external_system=external_system,
                metadata_ingestion_date=METADATA_INGESTION_DATE,
                metadata_ingestion_version="1",
                science_product_type=EXECUTION_BLOCK,
            ),
            ScienceProduct(
                external_name="uid___A002_Xcd8029_Xa892",
                filegroup_id=212043,
                external_system=external_system,
                metadata_ingestion_date=METADATA_INGESTION_DATE,
                metadata_ingestion_version="1",
                science_product_type=EXECUTION_BLOCK,
            ),
        ]

    def build_basic_ms(self) -> tuple:
        """ Returns basic measurement set metadata """

        ms_info = {
            "uid___A002_Xcd8029_Xa892": {
                "uid___A002_Xcd8029_Xa892.400887655.tar.gz": 504369560,
                PIPELINE_MANIFEST: 804,
                CASA_PIPESCRIPT: 680,
                PPR_FILENAME: 4584,
                WEBLOG: 1592335,
                MS_PP_REQUEST: 4584,
                AUX_PRODUCTS: 839,
                CASA_LOG: 1729,
                PIPELINE_AQUAREPORT: 3472,
                "uid___A002_Xcd8029_Xa892.ms.tgz": 519234359,
            },
            "uid___A002_Xcd8029_Xb0a4": {
                "uid___A002_Xcd8029_Xb0a4.400887655.tar.gz": 503501127,
                PIPELINE_MANIFEST: 804,
                CASA_LOG: 1729,
                WEBLOG: 1576475,
                CASA_PIPESCRIPT: 680,
                PIPELINE_AQUAREPORT: 3465,
                "uid___A002_Xcd8029_Xb0a4.ms.tgz": 518582184,
                AUX_PRODUCTS: 844,
                PPR_FILENAME: 4584,
                MS_PP_REQUEST: 4584,
            },
            "uid___A002_Xcd8029_Xfdd": {
                "uid___A002_Xcd8029_Xfdd.400887655.tar.gz": 489883578,
                PIPELINE_MANIFEST: 802,
                PIPELINE_AQUAREPORT: 3460,
                PPR_FILENAME: 4582,
                MS_PP_REQUEST: 4582,
                WEBLOG: 1562579,
                "uid___A002_Xcd8029_Xfdd.ms.tgz": 504128830,
                CASA_LOG: 1723,
                AUX_PRODUCTS: 836,
                CASA_PIPESCRIPT: 679,
            },
        }

        total_size = 0
        for mous in ms_info.keys():
            for _, size in ms_info[mous].items():
                total_size += size

        return ms_info, total_size

    def build_restored_ms(self):
        """ Returns restored measurement set metadata """
        ms_info = {
            "uid___A002_Xcd8029_Xb0a4": {
                "uid___A002_Xcd8029_Xb0a4.ms.tgz": 1325422495,
                "uid___A002_Xcd8029_Xa892.ms.tgz": 1331336708,
                "uid___A002_Xcd8029_Xfdd.ms.tgz": 1290775660,
                PPR_FILENAME: 5158,
            },
        }
        total_size = 0
        for _, size in ms_info["uid___A002_Xcd8029_Xb0a4"].items():
            total_size += size
        return ms_info, total_size

    def _compute_rawdata_size(self):
        total_size = 0
        for file_type in self.file_info.keys():
            metadata = self.file_info[file_type]
            for item in metadata.items():
                total_size += item[1]
        return total_size


class AlmaProprietaryImageProductSet:
    """ proprietary project 2019.1.00914.S, with image data """

    def __init__(self):
        self.project_code = "2019.1.00914.S"
        self.file_info, self.total_size = self._build_file_info()

    def _build_file_info(self) -> tuple:
        """
        TODO
        :return:
        """

        img_files_metadata = [
            get_file_info_from_json_file("ALMA_CONT_IMG_4d1b66da.json"),
            get_file_info_from_json_file("ALMA_CONT_IMG_71595054.json"),
            get_file_info_from_json_file("ALMA_IMG_CUBE.json"),
        ]

        file_info = {}
        total_size = 0
        for metadata in img_files_metadata:
            files = metadata["files"]
            for file in files:
                file_info[file["ngas_file_id"]] = file["size"]
                total_size += file["size"]
        return file_info, total_size


class AlmaAudiProductSet:
    """ AUDI products for project 2019.1.00914.S """

    def __init__(self):
        self.science_products = self.build_science_products()
        self.file_info = get_file_info_from_json_file("A001_X1296_Xa93_RAW.json")
        self.total_size = self._compute_total_size()

    def _compute_total_size(self):
        total_size = 0
        for file_type in self.file_info.keys():
            metadata = self.file_info[file_type]
            for item in metadata.items():
                total_size += item[1]
        return total_size

    def build_science_products(self) -> list:
        """ Returns 2019.1.00914.S science product in list """

        return [
            ScienceProduct(
                external_name="uid___A002_Xe5aacf_Xdde3",
                filegroup_id=398280,
                science_product_type=EXECUTION_BLOCK,
            )
        ]
