""" Test data: GBT download products """

import datetime
import warnings

from sqlalchemy import exc as sa_exc

from schema import Author, ExecutionBlock, Project, ScienceProduct
from shared.workspaces.test.test_data.utilities import (
    DATE_FORMAT,
    DATETIME_FORMAT,
    get_exec_block_details_from_loc_report,
)

# pylint: disable=C0301, R0201

class GbtProductSet:
    """ Encapsulates a GBT project and its products """

    def __init__(self):
        self.project = self.build_project()
        self.exec_blocks = self.build_exec_blocks()
        self.science_products = self.build_science_products()
        self.file_info, self.total_size = get_exec_block_details_from_loc_report(
            'AGBT17B_044_02', self.exec_blocks
        )

    def build_project(self) -> Project:
        """ Returns AGBT17B_044 metadata """

        with warnings.catch_warnings():
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            project = Project(project_code='AGBT17B_044',
                              starttime=58089.1619444443,
                              endtime=58138.2070370372,
                              proprietary_duration=365,
                              total_observation_time=27.055,
                              last_addition=datetime.datetime.strptime(
                                  '2020-10-21', DATE_FORMAT).date(),
                              title='Infall and Fragmentation of Cores in the Taurus: I Finding Cores with Infall',
                              abstract=' We propose to conduct a survey of investigating infall candidates and fragmentation toward a complete population of prestellar cores in Taurus in HCN, HCO+, N2H+, and NH2D. The proposed observation will trace infall motions within prestellar cores from 0.05 pc down to 0.0075 pc (1500 AU) spatial scale, which is a typical size scale of individual planetary systems. The scientific goal is to estimate a fraction of infall candidates out of a complete population of prestellar cores and to understand internal velocity structure during final gravitational collapse before forming stars. We shall also search for fragmentation within prestellar cores, and to investigate kinematics involved with fragmentation processes. We propose to conduct pointing observation toward 57 prestellar cores using the ARGUS of the GBT in the 89 GHz transitions of HCN and HCO+, 93 GHz transition of N2H+, and to map 2 prestellar cores, which are found to have infall motions, in N2H+ and NH2D.'
                              )
            project.authors = self.build_authors()
            return project

    def build_authors(self) -> list:
        """ Returns AGBT17B_044 authors' metadata """

        return [
            Author(
                username='seo3919',
                firstname='Youngmin',
                lastname='Seo',
                is_pi=True
            ),
            Author(
                username='schurch',
                firstname='Sarah',
                lastname='Church',
                is_pi=False
            ),
            Author(
                username='Paul Goldsmith',
                firstname='Paul',
                lastname='Goldsmith',
                is_pi=False
            ),
            Author(
                username='yancyshirley',
                firstname='Yancy',
                lastname='Shirley',
                is_pi=False
            ),
            Author(
                username='Michael Dunham',
                firstname='Michael',
                lastname='Dunham',
                is_pi=False
            ),
            Author(
                username='DTFrayer',
                firstname='Dave',
                lastname='Frayer',
                is_pi=False
            ),
        ]

    def build_exec_blocks(self):
        """ Returns AGBT17B_044's execution blocks """

        exec_blocks = [
            ExecutionBlock(execution_block_id=152679,
                           filegroup_id=553493,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
            ExecutionBlock(execution_block_id=152680,
                           filegroup_id=553494,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
            ExecutionBlock(execution_block_id=152681,
                           filegroup_id=553495,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
            ExecutionBlock(execution_block_id=152682,
                           filegroup_id=553496,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
            ExecutionBlock(execution_block_id=152683,
                           filegroup_id=553497,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
            ExecutionBlock(execution_block_id=152684,
                           filegroup_id=553498,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
            ExecutionBlock(execution_block_id=152685,
                           filegroup_id=553499,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
            ExecutionBlock(execution_block_id=152686,
                           filegroup_id=553500,
                           calibration_level='0',
                           calibration_status='GBT/NOCAL',
                           band_code='N/A'
                           ),
        ]
        return exec_blocks

    def build_science_products(self):
        """ Returns AGBT17B_044's science product
            (as a list, for consistency with other ProductSets)
        """

        return [
            ScienceProduct(
                filegroup_id=553493,
                external_system='GBT',
                science_product_type='dummy1',
                metadata_ingestion_date=datetime.datetime.strptime(
                    '2020-10-21 15:36:37.876148', DATETIME_FORMAT),
                metadata_ingestion_version='1',
                external_name='EXTERNAL NAME 2020-10-21 15:36:37.903180',
            )
        ]
